import pandas as pd
import numpy as np


def agent_attribution(tokens_df):
    tokens_df['char_att_agent'] = -1

    # condition_1 - direct verb subject
    subject_rows = tokens_df[(tokens_df["dependency_relation"] == "nsubj")
                             & (tokens_df["is_mention_head"] == 1)]
    verb_rows = tokens_df[(tokens_df["POS_tag"] == "VERB")
                          & (tokens_df["token_ID_within_document"].isin(subject_rows['syntactic_head_ID']))
                          ]

    merged_df = pd.merge(verb_rows, subject_rows, left_on='token_ID_within_document', right_on='syntactic_head_ID',
                         how='left')
    mention_head_id_list, verb_agent_head_id_list = merged_df['token_ID_within_document_y'].tolist(), merged_df[
        'token_ID_within_document_x'].tolist()
    del merged_df
    tokens_df.loc[verb_agent_head_id_list, 'char_att_agent'] = mention_head_id_list

    # condition_2 - conj synthactic relation
    verb_rows = tokens_df[(tokens_df["POS_tag"] == "VERB")
                          & (tokens_df["token_ID_within_document"].isin(subject_rows['syntactic_head_ID'].tolist()))
                          ]
    conj_verb_rows = tokens_df[(tokens_df["POS_tag"] == "VERB")
                               & (tokens_df["dependency_relation"] == "conj")
                               & (tokens_df["char_att_agent"] == -1)
                               & (tokens_df["syntactic_head_ID"].isin(verb_rows['token_ID_within_document']))]

    merged_df = pd.merge(conj_verb_rows, verb_rows, left_on='syntactic_head_ID', right_on='token_ID_within_document',
                         how='left')
    mention_head_id_list, verb_agent_head_id_list = merged_df['char_att_agent_y'].tolist(), merged_df[
        'token_ID_within_document_x'].tolist()
    del merged_df
    tokens_df.loc[verb_agent_head_id_list, 'char_att_agent'] = mention_head_id_list

    # fill and ensure the column is of integer type
    tokens_df['char_att_agent'] = tokens_df['char_att_agent'].fillna(-1).astype(int)

    return tokens_df

def patient_attribution(tokens_df):
    tokens_df['char_att_patient'] = -1

    verbs_rows_ids = tokens_df[(tokens_df['POS_tag'] == 'VERB')
    ]["token_ID_within_document"].tolist()

    mention_head_object_row = tokens_df[(tokens_df['dependency_relation'] == 'obj')
                                        & (tokens_df['syntactic_head_ID'].isin(verbs_rows_ids))
                                        & (tokens_df['is_mention_head'] == 1)]

    tokens_df.loc[mention_head_object_row['syntactic_head_ID'].tolist(), 'char_att_patient'] = mention_head_object_row[
        'token_ID_within_document'].tolist()

    # fill and ensure the column is of integer type
    tokens_df['char_att_patient'] = tokens_df['char_att_patient'].fillna(-1).astype(int)

    return tokens_df

def modifiers_attribution(tokens_df):
    tokens_df['char_att_mod'] = -1

    # Create a dictionary mapping token IDs to POS tags
    id_to_pos = dict(zip(tokens_df['token_ID_within_document'], tokens_df['syntactic_head_ID']))
    tokens_df['syntactic_head_head_id'] = tokens_df['syntactic_head_ID'].map(id_to_pos)

    is_mention_head_ids = tokens_df[tokens_df['is_mention_head'] == 1]["token_ID_within_document"].tolist()
    mention_adjectives = tokens_df[(tokens_df['POS_tag'] == "ADJ")
                                   & (tokens_df['dependency_relation'] != "flat:name")
                                   & (tokens_df['syntactic_head_ID'].isin(is_mention_head_ids))
                                   ]
    mention_adjectives_ids = mention_adjectives.index.tolist()
    tokens_df.loc[mention_adjectives_ids, 'char_att_mod'] = tokens_df['syntactic_head_ID']
    # ---
    mention_adjectives = tokens_df[(tokens_df['POS_tag'] == "ADJ")
                                   & (tokens_df['syntactic_head_ID'].isin(mention_adjectives_ids))
                                   ]
    mention_adjectives_ids = mention_adjectives.index.tolist()
    tokens_df.loc[mention_adjectives_ids, 'char_att_mod'] = tokens_df['syntactic_head_head_id']
    # ---
    adj_rows_id = tokens_df[(tokens_df['POS_tag'] == 'ADJ')
                            #                         &(tokens_df['mention_adj_head_id'] == 0)
                            & (tokens_df['dependency_relation'] == 'amod')
                            ]["token_ID_within_document"].tolist()
    mention_head_where_syntactic_head_is_adj = tokens_df[(tokens_df['is_mention_head'] == 1)
                                                         & (tokens_df['dependency_relation'].isin(['advmod']))
                                                         & (tokens_df['syntactic_head_ID'].isin(adj_rows_id))
                                                         ]
    mention_head_id_list, adj_id_list = mention_head_where_syntactic_head_is_adj['token_ID_within_document'].tolist(), \
    mention_head_where_syntactic_head_is_adj['syntactic_head_ID'].tolist()
    tokens_df.loc[adj_id_list, 'char_att_mod'] = mention_head_id_list

    # --- être
    être_head_ids = tokens_df[(tokens_df['lemma'] == 'être') & (tokens_df['POS_tag'] == 'AUX')][
        'syntactic_head_ID'].tolist()
    mention_head_ids = tokens_df[(tokens_df['is_mention_head'] == 1)
                                 & (tokens_df['dependency_relation'] == 'nsubj')
                                 ]
    adj_rows_id = tokens_df[(tokens_df['POS_tag'].isin(['ADJ', 'NOUN']))
                            #                         & (tokens_df['is_mention_head'] == 0)
                            & (tokens_df['token_ID_within_document'].isin(être_head_ids))
                            & (tokens_df['token_ID_within_document'].isin(
        mention_head_ids['syntactic_head_ID'].tolist()))
                            ].index.tolist()

    mention_head_with_adj = tokens_df[
        (tokens_df['token_ID_within_document'].isin(mention_head_ids['token_ID_within_document'].tolist()))
        & (tokens_df['syntactic_head_ID'].isin(adj_rows_id))]
    mention_head_id_list, adj_id_list = mention_head_with_adj['token_ID_within_document'].tolist(), \
    mention_head_with_adj['syntactic_head_ID'].tolist()
    tokens_df.loc[adj_id_list, 'char_att_mod'] = mention_head_id_list
    # --- conj other adjective

    mention_adj_ids = tokens_df[(tokens_df['char_att_mod'] != -1)]['token_ID_within_document'].tolist()
    conj_adj_rows = tokens_df[(tokens_df['POS_tag'] == 'ADJ')
                              & (tokens_df['char_att_mod'] == -1)
                              & (tokens_df['syntactic_head_ID'].isin(mention_adj_ids))
                              ]

    merged_df = pd.merge(conj_adj_rows, tokens_df, left_on='syntactic_head_ID', right_on='token_ID_within_document',
                         how='left')
    conj_adj_id_list, mention_head_id_list = merged_df['token_ID_within_document_x'].tolist(), merged_df[
        'char_att_mod_y'].tolist()
    del merged_df

    tokens_df.loc[conj_adj_id_list, 'char_att_mod'] = mention_head_id_list

    # fill and ensure the column is of integer type
    tokens_df['char_att_mod'] = tokens_df['char_att_mod'].fillna(-1).astype(int)

    return tokens_df

def avoir_poss_attribution(tokens_df):
    tokens_df['avoir_object_mention_head_id'] = -1

    avoir_ids = tokens_df[(tokens_df["lemma"] == 'avoir') & (tokens_df["POS_tag"] == 'VERB')][
        "token_ID_within_document"].tolist()
    avoir_mention_head_subject_ids = tokens_df[(tokens_df['dependency_relation'] == "nsubj")
                                               & (tokens_df['syntactic_head_ID'].isin(avoir_ids))
                                               & (tokens_df['is_mention_head'] == 1)
                                               ]
    avoir_with_object_ids = tokens_df[(tokens_df['dependency_relation'] == "obj")
                                      & (tokens_df['POS_tag'].isin(['NOUN', 'ADJ']))
                                      & (tokens_df['syntactic_head_ID'].isin(avoir_ids))
                                      ]
    merged_df = pd.merge(avoir_with_object_ids, avoir_mention_head_subject_ids, left_on='syntactic_head_ID',
                         right_on='syntactic_head_ID', how='left')
    merged_df = merged_df[~merged_df["token_ID_within_document_y"].isna()]
    merged_df['token_ID_within_document_y'] = merged_df['token_ID_within_document_y'].astype(int)

    avoir_object_id_list, mention_head_id_list = merged_df['token_ID_within_document_x'].tolist(), merged_df[
        'token_ID_within_document_y'].tolist()
    del merged_df

    tokens_df.loc[avoir_object_id_list, 'avoir_object_mention_head_id'] = mention_head_id_list

    tokens_df['avoir_object_mention_head_id'] = tokens_df['avoir_object_mention_head_id'].fillna(-1)
    tokens_df['avoir_object_mention_head_id'] = tokens_df['avoir_object_mention_head_id'].astype(int)
    tokens_df[tokens_df["avoir_object_mention_head_id"] != -1]
    return tokens_df
def articles_poss_attribution(tokens_df):
    possessive_articles_list = ['mon', 'ton', 'son', 'ma', 'ta', 'sa', 'notre', 'votre', 'leur', 'mes', 'mes', 'tes',
                                'ses', 'nos', 'vos', 'leurs']
    tokens_df['possesive_head_id'] = -1

    possessive_articles_mention_head = tokens_df[(tokens_df["word"].str.lower().isin(possessive_articles_list))
                                                 & (tokens_df["is_mention_head"] == 1)
                                                 & (tokens_df["POS_tag"] == ('DET'))
                                                 ]

    possessive_article_object = tokens_df[
        (tokens_df["token_ID_within_document"].isin(possessive_articles_mention_head['syntactic_head_ID'].tolist()))
        & (tokens_df["POS_tag"].isin(['NOUN', 'PROPN']))
        ]

    merged_df = pd.merge(possessive_article_object, possessive_articles_mention_head,
                         left_on='token_ID_within_document', right_on='syntactic_head_ID', how='left')
    merged_df[['sentence_ID_x', 'token_ID_within_document_x', 'token_ID_within_document_y']]

    mention_head_id_list, poss_object_id_list = merged_df['token_ID_within_document_y'].tolist(), merged_df[
        'token_ID_within_document_x'].tolist()
    del merged_df

    tokens_df.loc[poss_object_id_list, 'possesive_head_id'] = mention_head_id_list
    tokens_df['possesive_head_id'] = tokens_df['possesive_head_id'].fillna(-1)
    tokens_df['possesive_head_id'] = tokens_df['possesive_head_id'].astype(int)

    len(tokens_df[tokens_df['possesive_head_id'] != -1])
    tokens_df[tokens_df['possesive_head_id'] != -1]

    return tokens_df
def nmod_poss_attribution(tokens_df):
    tokens_df['nmod_possesive_head_id'] = -1
    de_rows = tokens_df[(tokens_df['lemma'] == 'de')]
    mention_head_nmod_rows = tokens_df[(tokens_df['is_mention_head'] == 1)
                                       & (tokens_df['dependency_relation'] == 'nmod')
                                       & (tokens_df['token_ID_within_document'].isin(
        de_rows['syntactic_head_ID'].tolist()))]
    possessiv_nominal_modifier = tokens_df[(tokens_df['POS_tag'] == 'NOUN')
                                           & (tokens_df['token_ID_within_document'].isin(
        mention_head_nmod_rows['syntactic_head_ID'].tolist()))]

    merged_df = pd.merge(possessiv_nominal_modifier, mention_head_nmod_rows, left_on='token_ID_within_document',
                         right_on='syntactic_head_ID', how='left')
    merged_df[['sentence_ID_x', 'token_ID_within_document_x', 'token_ID_within_document_y']]

    mention_head_id_list, poss_object_id_list = merged_df['token_ID_within_document_y'].tolist(), merged_df[
        'token_ID_within_document_x'].tolist()
    del merged_df

    tokens_df.loc[poss_object_id_list, 'nmod_possesive_head_id'] = mention_head_id_list
    tokens_df['nmod_possesive_head_id'] = tokens_df['nmod_possesive_head_id'].fillna(-1)
    tokens_df['nmod_possesive_head_id'] = tokens_df['nmod_possesive_head_id'].astype(int)

    len(tokens_df[tokens_df['nmod_possesive_head_id'] != -1])
    tokens_df[tokens_df['nmod_possesive_head_id'] != -1]

    return tokens_df
def combine_poss_columns(tokens_df):
    # combine poss columns
    # Create the new column using np.where
    tokens_df['char_att_poss'] = np.where(
        tokens_df['possesive_head_id'] != -1,  # If possesive_head_id is not 0
        tokens_df['possesive_head_id'],  # Use possesive_head_id
        np.where(
            tokens_df['avoir_object_mention_head_id'] != -1,  # Else, if avoir_object_mention_head_id is not 0
            tokens_df['avoir_object_mention_head_id'],  # Use avoir_object_mention_head_id
            -1  # Otherwise, use 0
        )
    )

    tokens_df['char_att_poss'] = np.where(
        tokens_df['char_att_poss'] != -1,  # If possesive_head_id is not 0
        tokens_df['char_att_poss'],  # Use possesive_head_id
        np.where(
            tokens_df['nmod_possesive_head_id'] != -1,  # Else, if avoir_object_mention_head_id is not 0
            tokens_df['nmod_possesive_head_id'],  # Use avoir_object_mention_head_id
            -1  # Otherwise, use 0
        )
    )
    tokens_df = tokens_df.drop(columns=['avoir_object_mention_head_id', 'possesive_head_id', 'nmod_possesive_head_id'])
    return tokens_df

def possessive_attribution(tokens_df):
    tokens_df = avoir_poss_attribution(tokens_df)
    tokens_df = articles_poss_attribution(tokens_df)
    tokens_df = nmod_poss_attribution(tokens_df)

    tokens_df = combine_poss_columns(tokens_df)

    if 'syntactic_head_head_id' in tokens_df.columns:
        tokens_df = tokens_df.drop(columns=['syntactic_head_head_id'])

    return tokens_df

def extract_attributes(entities_df, tokens_df):
    head_tokens_ids = entities_df['head_id'].tolist()
    tokens_df['is_mention_head'] = 0
    tokens_df.loc[head_tokens_ids, 'is_mention_head'] = 1

    tokens_df = agent_attribution(tokens_df)
    tokens_df = patient_attribution(tokens_df)
    tokens_df = modifiers_attribution(tokens_df)
    tokens_df = possessive_attribution(tokens_df)

    return tokens_df


import os
import csv
import re
import pandas as pd
import json

# Book_NLP_fr - Basic Loading and Saving functions
def load_sacr_file(file_name, files_directory="", extension=".sacr"):
    if not file_name.endswith(extension):
        file_name = file_name + extension

    sacr_file_path = os.path.join(files_directory, file_name)
    with open(sacr_file_path, 'r', encoding='utf-8') as file:
        sacr_content = file.read()  # Read the entire content of the file
    return sacr_content

def load_text_file(file_name, files_directory="", extension=".txt"):
    if not file_name.endswith(extension):
        file_name = file_name + extension

    text_file_path = os.path.join(files_directory, file_name)
    with open(text_file_path, 'r', encoding='utf-8') as file:
        text_content = file.read()  # Read the entire content of the file
    return text_content
def save_text_file(text_content, file_name, files_directory="", extension=".txt"):
    if not file_name.endswith(extension):
        file_name = file_name + extension

    # Ensure the directory exists; if not, create it
    if files_directory and not os.path.exists(files_directory):
        os.makedirs(files_directory)

    text_file_path = os.path.join(files_directory, file_name)

    with open(text_file_path, 'w', encoding='utf-8') as file:
        file.write(text_content)  # Write the text content to the file

def load_tokens_df(file_name, files_directory="", extension=".tokens"):

    if not file_name.endswith(extension):
        file_name = file_name + extension

    tokens_file_path = os.path.join(files_directory, file_name)
    tokens_df = pd.read_csv(tokens_file_path, delimiter='\t', quoting=csv.QUOTE_MINIMAL, keep_default_na=False)
    return tokens_df
def save_tokens_df(tokens_df, file_name, files_directory="", extension=".tokens"):
    # Check if the directory exists, if not, create it
    if not os.path.exists(files_directory):
        os.makedirs(files_directory)
        print(f"Directory '{files_directory}' created.")

    if not file_name.endswith(extension):
        file_name = file_name + extension
    tokens_file_path = os.path.join(files_directory, file_name)

    # Save the DataFrame as a .tokens file
    tokens_df.to_csv(tokens_file_path, sep='\t', index=False, quoting=csv.QUOTE_MINIMAL)

def load_entities_df(file_name, files_directory="", extension=".entities"):

    if not file_name.endswith(extension):
        file_name = file_name + extension

    entities_df_path = os.path.join(files_directory, file_name)
    entities_df = pd.read_csv(entities_df_path, delimiter='\t', quoting=csv.QUOTE_MINIMAL,  keep_default_na=False)
    return entities_df
def save_entities_df(entities_df, file_name, files_directory="", extension=".entities"):
    # Check if the directory exists, if not, create it
    if not os.path.exists(files_directory):
        os.makedirs(files_directory)
        print(f"Directory '{files_directory}' created.")

    if not file_name.endswith(extension):
        file_name = file_name + extension
    entities_file_path = os.path.join(files_directory, file_name)

    # Save the DataFrame as a .tokens file
    entities_df.to_csv(entities_file_path, sep='\t', index=False, quoting=csv.QUOTE_MINIMAL)

def clean_text(raw_text):
    raw_text = re.sub(r'�', ' ', raw_text)
    raw_text = re.sub(r'■', ' ', raw_text)
    raw_text = re.sub(r'•', ' ', raw_text)
    raw_text = raw_text.replace("”", '"')
    raw_text = raw_text.replace("“", '"')
    raw_text = raw_text.replace("’", "'")
    raw_text = raw_text.replace("' ", "'")
    raw_text = raw_text.replace(" , ", ", ")
    raw_text = raw_text.replace("\xa0", " ")
    # Replace multiple spaces (but not newlines) with a single space
    raw_text = re.sub(r'(?<=\S) {2,}(?=\S)', ' ', raw_text)
    raw_text = re.sub(r'[–—―‒]', '-', raw_text)
    raw_text = raw_text.replace(".-", ". -")
    raw_text = raw_text.replace("!-", "! -")
    raw_text = raw_text.replace("?-", "? -")
    return raw_text

def load_book_file(file_name, files_directory="", extension=".book"):
    if not file_name.endswith(extension):
        file_name = file_name + extension

    book_file_path = os.path.join(files_directory, file_name)
    with open(book_file_path, "r", encoding="utf-8") as f:
        characters_dict = json.load(f)
    return characters_dict
def save_book_file(characters_dict, file_name, files_directory="", extension=".book"):
    if not file_name.endswith(extension):
        file_name = file_name + extension

    # Ensure the directory exists; if not, create it
    if files_directory and not os.path.exists(files_directory):
        os.makedirs(files_directory)

    book_file_path = os.path.join(files_directory, file_name)

    with open(book_file_path, "w", encoding="utf-8") as f:
        json.dump(characters_dict, f, ensure_ascii=False, indent=4)



from .propp_fr_number import assign_number_to_PER_entities
from .propp_fr_gender import assign_gender_to_PER_entities

import pkg_resources
import numpy as np
from tqdm.auto import tqdm
import pandas as pd
from collections import Counter

def add_mention_number_and_gender_infos(entities_df):
    # Locate the CSV file within the package
    insee_path = pkg_resources.resource_filename('propp_fr', 'data/insee_names_fr_1900_2023.csv')
    # print("Insee_data Loaded successfully.")

    entities_df['number'] = assign_number_to_PER_entities(entities_df)['number']
    entities_df['gender'] = assign_gender_to_PER_entities(entities_df, insee_path=insee_path)['gender']

    not_PER_entities = entities_df[entities_df['cat'] != "PER"].copy()
    entities_df.loc[not_PER_entities.index, 'number'] = 'Not_Assigned'
    entities_df.loc[not_PER_entities.index, 'gender'] = 'Not_Assigned'

    return entities_df

## General entities_df functions
def add_mention_paragraphe_and_sentence_infos(entities_df, tokens_df):
    start_token_ids = entities_df['start_token'].tolist()
    entities_df['paragraph_ID'] = tokens_df.loc[start_token_ids, 'paragraph_ID'].tolist()
    entities_df['sentence_ID'] = tokens_df.loc[start_token_ids, 'sentence_ID'].tolist()
    entities_df['start_token_ID_within_sentence'] = tokens_df.loc[start_token_ids, 'token_ID_within_sentence'].tolist()
    return entities_df
    
def get_inner_to_outer_nesting_level(entities_df, start_column="start_token", end_column="end_token"):
    if "mention_len" not in entities_df.columns:
        entities_df['mention_len'] = entities_df[end_column] + 1 - entities_df[start_column]
    # Step 1: Initialize levels to 0
    entities_df["in_to_out_nested_level"] = 0

    # Step 2: Filter non-monogram entities and sort by mention length (ascending)
    non_monogram_entities = entities_df[entities_df["mention_len"] > 1].sort_values("mention_len", ascending=True)
    non_monogram_entities["temp_index"] = non_monogram_entities.index

    # Step 3: Convert DataFrame columns to NumPy arrays for faster indexing
    start_tokens = entities_df[start_column].values
    end_tokens = entities_df[end_column].values
    mention_lengths = entities_df["mention_len"].values
    nested_levels = entities_df["in_to_out_nested_level"].values

    # Step 4: Iterate over non-monogram entities
    for temp_index, start_token, end_token, mention_len in non_monogram_entities[["temp_index", start_column, end_column, "mention_len"]].values:
        # Step 5: Filter candidates for potential nesting
        candidates_mask = (mention_lengths < mention_len) & \
                          (start_tokens >= start_token) & \
                          (end_tokens <= end_token)

        # Step 6: Update nesting level if candidates exist
        if candidates_mask.any():
            max_nested_level = nested_levels[candidates_mask].max() + 1
            nested_levels[temp_index] = max_nested_level

    # Step 7: Assign the updated levels back to the DataFrame
    entities_df["in_to_out_nested_level"] = nested_levels

    return entities_df
    
def get_outer_to_inner_nesting_level(entities_df):
    # Step 1: Sort entities by start_token ascending, then by end_token descending
    entities_df = entities_df.sort_values(by=['start_token', 'end_token'], ascending=[True, False]).reset_index(drop=True)

    # Step 2: Initialize nested levels
    nested_levels = np.zeros(len(entities_df), dtype=int)

    # Step 3: Track active ranges (sorted by end_token)
    active_ranges = []

    for i, (start, end) in enumerate(zip(entities_df['start_token'], entities_df['end_token'])):
        # Remove ranges from active_ranges that are no longer active (end < start of the current mention)
        active_ranges = [ar for ar in active_ranges if ar[1] >= start]

        # The current level is the size of active_ranges
        nested_levels[i] = len(active_ranges)

        # Add the current range to active_ranges
        active_ranges.append((start, end))

    # Step 4: Assign the calculated nested levels back to the DataFrame
    entities_df['out_to_in_nested_level'] = nested_levels

    return entities_df

def get_nested_entities_count(entities_df):
    # Step 1: Initialize count column
    entities_df["nested_entities_count"] = 0
    entities_df["temp_index"] = entities_df.index

    # Step 2: Extract values as NumPy arrays for speed
    start_tokens = entities_df["start_token"].values
    end_tokens = entities_df["end_token"].values
    mention_lengths = entities_df["mention_len"].values
    nested_counts = np.zeros(len(entities_df), dtype=int)

    # Step 3: Filter outer entities (which contain other entities)
    outer_mask = entities_df["in_to_out_nested_level"].values > 0
    outer_indices = np.where(outer_mask)[0]

    # Step 4: Iterate efficiently over outer entities
    for idx in outer_indices:
        start_token, end_token, mention_len = start_tokens[idx], end_tokens[idx], mention_lengths[idx]

        # Vectorized filtering for inner mentions
        inside_mask = (mention_lengths < mention_len) & \
                      (start_tokens >= start_token) & \
                      (end_tokens <= end_token)

        # Store the count of nested entities
        nested_counts[idx] = np.sum(inside_mask)

    # Step 5: Assign results back to DataFrame
    entities_df["nested_entities_count"] = nested_counts

    return entities_df

def assign_mention_head_id(entities_df, tokens_df):
    entities_df = entities_df.copy()
    entities_df['head_id'] = entities_df['start_token']

    filtered_entities_df = entities_df[entities_df['mention_len'] != 1]

    mention_head_ids = np.zeros(len(filtered_entities_df), dtype=int)

    for idx, (start_token, end_token) in enumerate(zip(filtered_entities_df['start_token'], filtered_entities_df['end_token'])):
        # Get mention tokens
        mention_token_df = tokens_df.loc[start_token:end_token].copy()
        if mention_token_df.empty:
            mention_head_ids[idx] = start_token  # Default to first token if no valid head is found
            continue

        # Check if head is inside the mention
        mention_token_df['head_is_inside_mention'] = mention_token_df['syntactic_head_ID'].isin(
            mention_token_df['token_ID_within_document']
        )

        # Prioritize heads inside the mention
        valid_heads = mention_token_df[mention_token_df['head_is_inside_mention']]['syntactic_head_ID']

        if not valid_heads.empty:
            # If multiple heads exist inside, choose the most frequent one
            mention_head_id = valid_heads.value_counts().idxmax()
        else:
            # If no valid head inside, force the first token of the mention
            mention_head_id = start_token

        mention_head_ids[idx] = mention_head_id

    # Assign back
    entities_df.loc[filtered_entities_df.index, "head_id"] = mention_head_ids

    return entities_df

def mention_head_syntactic_infos(entities_df, tokens_df):
    head_token_ids = entities_df['head_id'].tolist()
    head_tokens_rows = tokens_df.loc[head_token_ids, ['word', 'dependency_relation', 'syntactic_head_ID']]
    entities_df[['head_word', 'head_dependency_relation', 'head_syntactic_head_ID']] = head_tokens_rows.values.tolist()
    return entities_df
def assign_mention_prop(entities_df, tokens_df):
    # Define the mapping dictionary
    mapping_dict = {
        'PROPN': 'PROP',
        'PRON': 'PRON',
        'DET': 'PRON',
        'ADP': 'PRON',
        'PUNCT': 'PRON',
        'NOUN': 'NOM',
    }
    default_value = "NOM"

    entities_df['POS_tag'] = pd.merge(entities_df['head_id'], tokens_df, left_on='head_id', right_index=True)['POS_tag']
    # Use map to apply the mapping dictionary, setting unmapped values to NaN

    # Apply the mapping dictionary to the 'POS_tag' column
    entities_df['prop'] = entities_df['POS_tag'].map(mapping_dict).fillna(default_value)

    special_pronouns_tokens = ['moi', 'mien', 'miens', 'mienne', 'miennes', 'tien', 'tiens', 'tienne', 'tiennes',
                               'vôtre', 'vôtres', 'siens', 'siennes', 'sien']
    special_pronouns_tokens_id = list(tokens_df[tokens_df['word'].str.lower().isin(special_pronouns_tokens)].index)
    head_id_in_special_tokens = list(entities_df[entities_df['head_id'].isin(special_pronouns_tokens_id)].index)
    entities_df.loc[head_id_in_special_tokens, ['prop']] = "PRON"

    # Propagate PROP tag to all
    occurrence_treshold = 5
    proper_name_mentions = dict(Counter(entities_df[entities_df['prop'] == 'PROP']['text']))
    proper_name_mentions = [key for key in proper_name_mentions.keys() if
                            proper_name_mentions[key] >= occurrence_treshold]
    proper_rows = entities_df[entities_df['text'].isin(proper_name_mentions)]
    entities_df.loc[proper_rows.index, 'prop'] = 'PROP'

    always_pronouns_mentions = ['je', 'me', 'moi', "j'", "m'", 'mon', 'ma', 'mes', 'nous', 'notre', 'nos', 'moi - même', 'moi-même', 'mien', 'miens', 'mienne',
                                'miennes', 'nôtre', 'nous-mêmes', 'nous - mêmes', 'tu', 'toi', 'te', "t'", 'ton', 'ta', 'tes', 'vous', 'vôtre', 'vos', 'votre',
                                'tien', 'tiens', 'tienne', 'tiennes', 'vous - même', 'vous-même', 'toi-même', 'toi - même', 'vous-mêmes', 'vous - mêmes', 'il',
                                'elle', 'lui', 'son', 'sa', "l'", 'ses', 'le', 'la', 'se', 'ils', 'elles', 'leur', 'les', 'leurs', 'eux', "s'", 'elle-même',
                                'lui-même', 'sienne', 'sien', 'sienne', 'siennes','un', 'une', "l' autre", 'tous', 'celui-ci', 'duquel', 'celle', 'celui', 'qui',
                                'que', 'dont', "qu'", "eux - mêmes", "eux-mêmes", "elles - mêmes", "elles-mêmes"]
    pronoun_rows = entities_df[entities_df['text'].isin(always_pronouns_mentions)]
    entities_df.loc[pronoun_rows.index, 'prop'] = 'PRON'

    always_noun_mentions = ['madame', 'mademoiselle', "monsieur", "papa", "maman"]
    noun_rows = entities_df[entities_df['text'].isin(always_noun_mentions)]
    entities_df.loc[noun_rows.index, 'prop'] = 'NOM'

    return entities_df
def assign_grammatical_person(entities_df):
    grammatical_person_dict = {"1": ['je', 'me', 'moi', "j'", "m'", 'mon', 'ma', 'mes', 'nous', 'notre', 'nos', 'moi - même', 'moi-même', 'mien', 'miens', 'mienne','miennes', 'nôtre', 'nous-mêmes', 'nous - mêmes'],
                               "2": ['tu', 'toi', 'te', "t'", 'ton', 'ta', 'tes', 'vous', 'vôtre', 'vos', 'votre', 'tien', 'tiens', 'tienne', 'tiennes', 'vous - même', 'vous-même', 'toi-même', 'toi - même', 'vous-mêmes', 'vous - mêmes'],
                               "3": ['il', 'elle', 'lui', 'son', 'sa', "l'", 'ses', 'le', 'la', 'se', 'ils', 'elles', 'leur', 'les', 'leurs', 'eux', "s'", 'elle-même', 'lui-même', 'sienne', 'sien', 'sienne', 'siennes','un', 'une', "l' autre", 'tous', 'celui-ci', 'duquel', 'celle', 'celui'],
                               "4": ['qui', 'que', 'dont', "qu'"]}

    # Reverse the dictionary to map each word to its grammatical person
    word_to_person = {word: person for person, words in grammatical_person_dict.items() for word in words}

    # Map entities_df['text'].str.lower() to grammatical persons
    entities_df['grammatical_person'] = entities_df['text'].str.lower().map(word_to_person).fillna("3").astype(int)
    return entities_df

def harmonize_prefix(entities_df, replace_dict=None, text_column="text", ):
    if replace_dict == None:
        replace_dict = {"mme": "madame",
                        "m.": "monsieur",
                        "mgr": "monsieur",
                        "sieur": "monsieur",
                        "le sieur": "monsieur",
                        "du sieur": "monsieur",
                        "mm.": "messieurs",
                        "me": "maître",
                        "dr": "docteur",
                        "le dr": "docteur",
                        "du dr": "docteur",
                        "l' abbé": "abbé",
                        "le marquis": "marquis",
                        "la marquise": "marquise",
                        "le comte": "comte",
                        "du comte": "comte",
                        "la comtesse": "comtesse",
                        "le duc": "duc",
                        "du duc": "duc",
                        "le curé": "curé",
                        "la duchesse": "duchesse",
                        "la baronne": "baronne",
                        "le baron": "baron",
                        "du baron": "baron",
                        "au baron": "baron",
                        "le capitaine": "capitaine",
                        "du capitaine": "capitaine",
                        "le lieutenant": "lieutenant",
                        "du lieutenant": "lieutenant",
                        "le cardinal": "cardinal",
                        "le docteur": "docteur",
                        "du docteur": "docteur",
                        "du colonel": "colonel",
                        "le colonel": "colonel",
                        "le chevalier": "chevalier",
                        "mlle": "mademoiselle",
                        "l' ambassadeur": "ambassadeur",
                        "le sous-lieutenant": "sous-lieutenant",
                        "du sous-lieutenant": "sous-lieutenant",
                        "le père": "père",
                        "la mère": "mère",
                        "le roi": "roi",
                        "au roi": "roi",
                        "la reine": "reine",
                        "la donna": "donna",
                        "aux": "les",
                        "au": "le",
                        "ma cousine": "cousine",
                        "la famille": "famille",
                        "la société": "société",
                        "la maison": "maison",
                        },

    # Function to replace values from the dictionary
    def replace_from_dict(text):
        # Now replace other values from the dictionary
        for key, value in replace_dict.items():
            if text.startswith(key):
                text = text.replace(f"{key} ", f"{value} ").strip()
        return text

    # Apply the replacement function
    entities_df[text_column] = entities_df[text_column].fillna("").apply(replace_from_dict)
    return entities_df
def correct_PROP(entities_df, tokens_df):
    NOM_entities = entities_df.copy()
    NOM_entities["POS_tags"] = NOM_entities.apply(
        lambda row: " ".join(tokens_df.loc[row["start_token"]:row["end_token"], "POS_tag"]), axis=1)

    NOM_entities = NOM_entities[NOM_entities["prop"].isin(["NOM", "PRON"])].copy()
    NOM_entities = NOM_entities[NOM_entities["mention_len"] > 1]
    NOM_entities = NOM_entities[NOM_entities["POS_tags"].str.contains("PROPN")]
    replace_dict = {"mme": "madame",
                    "m.": "monsieur",
                    "mgr": "monsieur",
                    "sieur": "monsieur",
                    "le sieur": "monsieur",
                    "du sieur": "monsieur",
                    "mm.": "messieurs",
                    "me": "maître",
                    "dr": "docteur",
                    "le dr": "docteur",
                    "du dr": "docteur",
                    "l' abbé": "abbé",
                    "le marquis": "marquis",
                    "la marquise": "marquise",
                    "le comte": "comte",
                    "du comte": "comte",
                    "la comtesse": "comtesse",
                    "le duc": "duc",
                    "du duc": "duc",
                    "le curé": "curé",
                    "la duchesse": "duchesse",
                    "la baronne": "baronne",
                    "le baron": "baron",
                    "du baron": "baron",
                    "au baron": "baron",
                    "le capitaine": "capitaine",
                    "du capitaine": "capitaine",
                    "le lieutenant": "lieutenant",
                    "du lieutenant": "lieutenant",
                    "le cardinal": "cardinal",
                    "le docteur": "docteur",
                    "du docteur": "docteur",
                    "du colonel": "colonel",
                    "le colonel": "colonel",
                    "le chevalier": "chevalier",
                    "mlle": "mademoiselle",
                    "l' ambassadeur": "ambassadeur",
                    "le sous-lieutenant": "sous-lieutenant",
                    "du sous-lieutenant": "sous-lieutenant",
                    "le père": "père",
                    "la mère": "mère",
                    "le roi": "roi",
                    "au roi": "roi",
                    "la reine": "reine",
                    "la donna": "donna",
                    "aux": "les",
                    "au": "le",
                    "ma cousine": "cousine",
                    "la famille": "famille",
                    "la société": "société",
                    "la maison": "maison",
                    }
    NOM_entities = harmonize_prefix(NOM_entities, replace_dict=replace_dict)
    NOM_entities = NOM_entities[~NOM_entities["text"].str.startswith(
        ('mère de', 'père de', "la cousine d'", "la fille du", "la fille de", "des filles de"))]

    entities_df.loc[NOM_entities.index.tolist(), "prop"] = "PROP"
    return entities_df


def add_features_to_entities(entities_df, tokens_df):
    lower_texts = [text.lower() for text in entities_df['text'].tolist()]
    entities_df['text'] = lower_texts
    entities_df['mention_len'] = entities_df['end_token'] + 1 - entities_df['start_token']
    entities_df = add_mention_paragraphe_and_sentence_infos(entities_df, tokens_df)
    entities_df = get_outer_to_inner_nesting_level(entities_df)
    entities_df = get_inner_to_outer_nesting_level(entities_df)
    entities_df = get_nested_entities_count(entities_df)
    entities_df = assign_mention_head_id(entities_df, tokens_df)
    entities_df = mention_head_syntactic_infos(entities_df, tokens_df)
    entities_df = assign_mention_prop(entities_df, tokens_df)

    # Correct error of "prop" category
    entities_df = correct_PROP(entities_df, tokens_df)

    entities_df = add_mention_number_and_gender_infos(entities_df)
    entities_df = assign_grammatical_person(entities_df)

    mentions_text = []
    for start_token, end_token in entities_df[["start_token", "end_token"]].values:
        text = " ".join(list(tokens_df.loc[start_token:end_token, "word"]))
        mentions_text.append(text)
    entities_df["text"] = mentions_text
    entities_df["text"] = entities_df["text"].astype(str).str.lower()
    return entities_df



from transformers import AutoTokenizer, AutoModel
import torch
import numpy as np
from tqdm.auto import tqdm
import pandas as pd
import gc

def load_tokenizer_and_embedding_model(model_name="almanach/camembert-large"):

    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModel.from_pretrained(model_name, output_hidden_states=True)
    print(f"Tokenizer and Embedding Model Initialized: {model_name}")

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)

    return tokenizer, model

def tokenize_text(text, tokenizer):
    """Tokenizes text and retrieves token IDs with byte offsets"""
    encoding = tokenizer(text,
                         return_offsets_mapping=True,
                         return_tensors="pt",
                         return_attention_mask=False,
                         add_special_tokens=False)

    subword_ids = encoding["input_ids"].squeeze().tolist()  # Token IDs
    subword_offsets = encoding["offset_mapping"].squeeze().tolist()  # (Start, End) byte positions
    subwords = tokenizer.convert_ids_to_tokens(subword_ids)  # Convert IDs to readable tokens

    return subwords, subword_ids, subword_offsets

def find_subwords_for_tokens(tokens_df, subword_offsets):
    subword_onsets = np.array([offset[0] for offset in subword_offsets])
    subword_offsets = np.array([offset[1] for offset in subword_offsets])

    tokens_subword_ids = []

    for token_byte_onset, token_byte_offset in tokens_df[["byte_onset", "byte_offset"]].values:
        # Find the first subword that starts at or before the token onset
        start_idx = np.searchsorted(subword_onsets, token_byte_onset, side="right") - 1
        start_idx = max(start_idx, 0)  # Ensure we don't go below index 0

        # Expand right while ensuring subwords still overlap with the token
        end_idx = start_idx
        while end_idx < len(subword_offsets) and subword_onsets[end_idx] < token_byte_offset:
            end_idx += 1

        # Allow subwords that start before the token but overlap significantly
        contained_subwords_ids = [
            i for i in range(start_idx, end_idx)
            if not (subword_offsets[i] <= token_byte_onset or subword_onsets[i] >= token_byte_offset)
        ]

        tokens_subword_ids.append(contained_subwords_ids)

    return tokens_subword_ids

# def get_boudaries_list(tokens_df, sub_word_offsets, sliding_window_size=0, sliding_window_overlap=0.5):
#     token_offsets = np.array(tokens_df["byte_offset"].tolist())
#
#     # Convert to NumPy set for fast lookup
#     token_offset_set = set(token_offsets)
#
#     # Get indices of possible subword boundaries
#     possible_end_tokens = np.array([i for i, (_, end) in enumerate(sub_word_offsets) if end in token_offset_set])
#     possible_start_tokens = np.array([0] + (possible_end_tokens + 1).tolist())
#
#     overlapping_window_boundaries = []
#     window_start = 0
#
#     while window_start < len(sub_word_offsets):
#         if window_start + sliding_window_size >= len(sub_word_offsets):
#             window_end = len(sub_word_offsets)
#             # Append the window boundaries
#             overlapping_window_boundaries.append((window_start, window_end))
#             return overlapping_window_boundaries
#         # Find the rightmost valid `window_end` within range
#         max_idx = np.searchsorted(possible_end_tokens, window_start + sliding_window_size, side='right') - 1
#         window_end = possible_end_tokens[max_idx] if max_idx >= 0 else window_start
#
#         # Append the window boundaries
#         overlapping_window_boundaries.append((window_start, window_end))
#
#         # Shift the window with overlap
#         new_start_pos = window_end - int(sliding_window_size * sliding_window_overlap)
#
#         # Find the closest valid start token using binary search
#         start_idx = np.searchsorted(possible_start_tokens, new_start_pos, side='right') - 1
#         window_start = possible_start_tokens[start_idx] if start_idx >= 0 else 0
#
#     return overlapping_window_boundaries

def get_boudaries_list(tokens_df, sub_word_offsets, sliding_window_size=0, sliding_window_overlap=0.5):
    import numpy as np

    token_offsets = np.array(tokens_df["byte_offset"].tolist())
    token_offset_set = set(token_offsets)

    possible_end_tokens = np.array([i for i, (_, end) in enumerate(sub_word_offsets) if end in token_offset_set])
    possible_start_tokens = np.array([0] + (possible_end_tokens + 1).tolist())

    overlapping_window_boundaries = []
    window_start = 0
    prev_window_start = -1

    while window_start < len(sub_word_offsets):
        if window_start == prev_window_start:
            print(f"No progress detected at window_start={window_start} — breaking to avoid infinite loop.")
            break
        prev_window_start = window_start

        if window_start + sliding_window_size >= len(sub_word_offsets):
            window_end = len(sub_word_offsets)
            overlapping_window_boundaries.append((window_start, window_end))
            break

        max_idx = np.searchsorted(possible_end_tokens, window_start + sliding_window_size, side='right') - 1
        window_end = possible_end_tokens[max_idx] if max_idx >= 0 else window_start

        overlapping_window_boundaries.append((window_start, window_end))

        # Calculate next start with overlap
        new_start_pos = window_end - int(sliding_window_size * sliding_window_overlap)
        start_idx = np.searchsorted(possible_start_tokens, new_start_pos, side='right') - 1

        next_window_start = possible_start_tokens[start_idx] if start_idx >= 0 else window_start + 1

        # Force forward progress
        if next_window_start <= window_start:
            next_window_start = window_start + 1

        window_start = next_window_start

    return overlapping_window_boundaries

def compute_sub_word_embeddings(boundaries_list, token_ids, model, mini_batch_size=10, padding_token_id=0, sliding_window_size=0, device='cpu'):

    # Convert token_ids to a single tensor before the loop
    token_ids = torch.tensor(token_ids, dtype=torch.long, device=device)
    max_token_id = token_ids.size(0)

    # Precompute the padding tensor once
    padding_tensor = torch.tensor([padding_token_id] * sliding_window_size, dtype=torch.long).unsqueeze(0).to(device)

    batch_input_ids = []
    all_embeddings_batches = []

    with torch.no_grad():
        for start_boundary, end_boundary in tqdm(boundaries_list, desc='Embedding Tokens', leave=False):
            input_ids = token_ids[start_boundary:end_boundary]
            real_tokens_length = input_ids.size(0)

            # Pad only if needed
            if real_tokens_length < sliding_window_size:
                padded_input = torch.cat([input_ids.unsqueeze(0), padding_tensor[:, real_tokens_length:]], dim=1)
            else:
                padded_input = input_ids.unsqueeze(0)


            batch_input_ids.append(padded_input)

            # Process the batch based on the specified conditions
            if (len(batch_input_ids) == mini_batch_size) or (
                    end_boundary >= max_token_id):  # Ensure end_boundary is a single value
                # Process the batch
                batch_input_ids_tensor = torch.cat(batch_input_ids,
                                                   dim=0)  # Concatenate all input tensors along the batch dimension
                attention_mask = (batch_input_ids_tensor != padding_token_id).long()  # Create attention mask

                # Move to device if not already
                batch_input_ids_tensor = batch_input_ids_tensor.to(device)
                attention_mask = attention_mask.to(device)

                # Get model outputs
                outputs = model(batch_input_ids_tensor, attention_mask=attention_mask)
                last_hidden_states = outputs.last_hidden_state  # Get last hidden states

                # Reshape last hidden states
                last_hidden_states = last_hidden_states.view(-1, last_hidden_states.shape[2])

                # Flatten hidden states and filter out padding
                token_embeddings = last_hidden_states[attention_mask.view(-1) == 1]

                all_embeddings_batches.append(token_embeddings.cpu())

                # Reset the batch
                batch_input_ids = []

    # Concatenate all embeddings
    all_embeddings = torch.cat(all_embeddings_batches)

    subword_indices = []
    for start_boundary, end_boundary in boundaries_list:
        subword_indices += list(range(start_boundary, end_boundary))

    return subword_indices, all_embeddings

def average_embeddings_from_overlapping_sliding_windows(subword_indices, all_embeddings):
    subword_indices = torch.tensor(subword_indices, device=all_embeddings.device)  # Convert to tensor
    unique_indices, inverse_indices = torch.unique(subword_indices, return_inverse=True)  # Get unique indices

    # Get counts for each unique subword index
    counts = torch.bincount(inverse_indices, minlength=len(unique_indices)).float().unsqueeze(1)

    # Use bincount to sum embeddings efficiently
    summed_embeddings = torch.zeros((len(unique_indices), all_embeddings.shape[1]), device=all_embeddings.device)
    summed_embeddings.index_add_(0, inverse_indices, all_embeddings)

    # Compute mean embeddings
    mean_embeddings = summed_embeddings / counts

    return mean_embeddings

def get_token_embeddings(mean_subword_embeddings,
                         tokens_subword_ids,
                         subword_pooling_strategy="average" # ["average", "first", "last", "first_last"]
                         ):

    """
    Pool subword embeddings to get token-level embeddings, with handling for empty token-subword lists.

    Args:
        mean_subword_embeddings (torch.Tensor): Tensor containing subword embeddings (num_subwords, embedding_dim).
        tokens_subword_ids (list of lists): List of lists where each list contains the indices of the subwords
                                            corresponding to a token.
        subword_pooling_strategy (str): The pooling strategy to use for merging subword embeddings. Options are:
                                       "average", "first", "last", "first_last".

    Returns:
        torch.Tensor: Token-level embeddings with shape (len(tokens_subword_ids), embedding_dim)
    """

    # Pre-allocate zero tensor using the shape of the first subword embedding
    zero_tensor = torch.zeros(mean_subword_embeddings.shape[1])

    # Initialize an empty tensor for token embeddings (same length as number of tokens)
    token_embeddings = torch.zeros(len(tokens_subword_ids), mean_subword_embeddings.shape[1])

    for idx, token_subword_ids in enumerate(tokens_subword_ids):
        if len(token_subword_ids) == 0:
            token_embeddings[idx] = zero_tensor
        else:
            # Extract the subword embeddings corresponding to the current token
            token_subword_embeddings = mean_subword_embeddings[token_subword_ids]

            # Apply pooling strategy
            if subword_pooling_strategy == "average":
                # Average the embeddings of all subwords
                token_embeddings[idx] = token_subword_embeddings.mean(dim=0)
            elif subword_pooling_strategy == "first":
                # Take the first subword embedding
                token_embeddings[idx] = token_subword_embeddings[0]
            elif subword_pooling_strategy == "last":
                # Take the last subword embedding
                token_embeddings[idx] = token_subword_embeddings[-1]
            elif subword_pooling_strategy == "first_last":
                # Average the first and last subword embeddings
                token_embeddings[idx] = (token_subword_embeddings[0] + token_subword_embeddings[-1]) / 2
            elif subword_pooling_strategy == "max":
            	# Take the maximum value across subwords
            	token_embeddings[idx] = token_subword_embeddings.max(dim=0)[0]
            else:
                raise ValueError(f"Unknown pooling strategy: {subword_pooling_strategy}")

    return token_embeddings

def get_embedding_tensor_from_tokens_df(text,
                                        tokens_df,
                                        tokenizer,
                                        model,
                                        sliding_window_size='max',
                                        mini_batch_size=12,
                                        sliding_window_overlap=0.5,
                                        subword_pooling_strategy="average", # ["average", "first", "last", "first_last"]
                                        device=None):

    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if sliding_window_size=='max':
        if tokenizer.model_max_length > 100000:
            sliding_window_size = 512
        else:
            sliding_window_size = tokenizer.model_max_length


    padding_token_id = int(tokenizer.pad_token_id)
    subwords, subword_ids, subword_offsets = tokenize_text(text, tokenizer)
    tokens_subword_ids = find_subwords_for_tokens(tokens_df, subword_offsets)
    overlapping_window_boundaries = get_boudaries_list(tokens_df, sub_word_offsets=subword_offsets, sliding_window_size=sliding_window_size, sliding_window_overlap=sliding_window_overlap)
    subword_indices, all_subword_embeddings = compute_sub_word_embeddings(overlapping_window_boundaries, subword_ids, model, mini_batch_size=mini_batch_size, padding_token_id=padding_token_id, sliding_window_size=sliding_window_size, device=device)
    mean_subword_embeddings = average_embeddings_from_overlapping_sliding_windows(subword_indices, all_subword_embeddings)
    del all_subword_embeddings
    gc.collect()
    tokens_embeddings_tensor = get_token_embeddings(mean_subword_embeddings,
                                                    tokens_subword_ids,
                                                    subword_pooling_strategy=subword_pooling_strategy # ["average", "first", "last", "first_last"]
                             )

    # tokens_df["subword_ids"] = tokens_subword_ids
    # tokens_df["subword_offsets"] = tokens_df["subword_ids"].apply(lambda subword_ids: [subword_offsets[i] for i in subword_ids])
    # tokens_df["subwords"] = tokens_df["subword_ids"].apply(lambda subword_ids: [subwords[i] for i in subword_ids])


    return tokens_embeddings_tensor


# from transformers import AutoTokenizer, AutoModel
# import torch
# import numpy as np
# from tqdm.auto import tqdm
# import pandas as pd
#
# ## Get BERT embeddings from tokens_df
# def fast_tokennize_tokens_df(tokens_df, tokenizer):
#     tokens_dict = []
#
#     # List of words from the dataframe
#     words = tokens_df["word"].astype(str).tolist()  # Ensure all words are strings
#     token_ids = tokens_df["token_ID_within_document"].tolist()
#
#     # Tokenize all words at once using the fast tokenizer
#     batch_encoding = tokenizer.batch_encode_plus(words, add_special_tokens=False)
#
#     # Preallocate the length to reduce append overhead
#     for token_id, word_subwords in zip(token_ids, batch_encoding['input_ids']):
#         tokens_dict.extend([{
#             "token_id": token_id,
#             "camembert_token_id": camembert_token_id
#         } for camembert_token_id in word_subwords])
#
#     return tokens_dict
# def get_boudaries_list(max_token_id=0, sliding_window_size=0, sliding_window_overlap=0.5):
#     # Parameters
#     sliding_window_overlap = int(sliding_window_size * sliding_window_overlap)  # 50% overlap
#     min_token_id = 0
#
#     # Create the sliding window boundaries efficiently using NumPy
#     # Step size accounts for the overlap, reducing redundant token IDs
#     step_size = sliding_window_size - sliding_window_overlap
#
#     # Create an array of starting boundaries
#     min_boundaries = np.arange(min_token_id, max_token_id, step_size)
#
#     # Create corresponding max boundaries (ensure they don't exceed max_token_id)
#     max_boundaries = np.minimum(min_boundaries + sliding_window_size, max_token_id)
#
#     # Combine min and max boundaries into a list of tuples and convert to int
#     boundaries_list = [(int(min_boundary), int(max_boundary)) for min_boundary, max_boundary in
#                        zip(min_boundaries, max_boundaries)]
#
#     return boundaries_list
# def compute_sub_word_embeddings(boundaries_list, tokens_dict, model, mini_batch_size=100, padding_token_id=0, sliding_window_size=0, device='cpu'):
#
#     max_token_id = len(tokens_dict)  # You may want to adjust this based on your logic
#
#     tokens_dict_df = pd.DataFrame(tokens_dict)
#
#     # Precompute the padding tensor once
#     padding_tensor = torch.tensor([padding_token_id] * sliding_window_size, dtype=torch.long).unsqueeze(0).to(device)
#
#     batch_input_ids = []
#     all_embeddings_batches = []
#
#     with torch.no_grad():
#         for start_boundary, end_boundary in tqdm(boundaries_list, desc='Embedding Tokens', leave=False):
#             input_ids = tokens_dict_df[start_boundary:end_boundary]['camembert_token_id'].tolist()
#             real_tokens_length = len(input_ids)
#
#             # Convert to tensor and send to device
#             input_tensor = torch.tensor(input_ids, dtype=torch.long).unsqueeze(0).to(device)
#
#             # Check if padding is needed
#             if real_tokens_length < sliding_window_size:
#                 # Pad the tensor to the right size
#                 padded_input = torch.cat([input_tensor, padding_tensor[:, real_tokens_length:]], dim=1)
#             else:
#                 padded_input = input_tensor
#
#             batch_input_ids.append(padded_input)
#
#             # Process the batch based on the specified conditions
#             if (len(batch_input_ids) == mini_batch_size) or (
#                     end_boundary == max_token_id):  # Ensure end_boundary is a single value
#                 # Process the batch
#                 batch_input_ids_tensor = torch.cat(batch_input_ids,
#                                                    dim=0)  # Concatenate all input tensors along the batch dimension
#                 attention_mask = (batch_input_ids_tensor != padding_token_id).long()  # Create attention mask
#
#                 # Move to device if not already
#                 batch_input_ids_tensor = batch_input_ids_tensor.to(device)
#                 attention_mask = attention_mask.to(device)
#
#                 # Get model outputs
#                 outputs = model(batch_input_ids_tensor, attention_mask=attention_mask)
#                 last_hidden_states = outputs.last_hidden_state  # Get last hidden states
#
#                 # Reshape last hidden states
#                 last_hidden_states = last_hidden_states.view(-1, last_hidden_states.shape[2])
#
#                 embedding_end_index = real_tokens_length - sliding_window_size
#                 if embedding_end_index != 0:
#                     token_embeddings = last_hidden_states[:embedding_end_index, :]  # Shape: [num_tokens, 1024]
#                 else:
#                     token_embeddings = last_hidden_states
#
#                 all_embeddings_batches.append(token_embeddings.cpu())
#
#                 # Reset the batch
#                 batch_input_ids = []
#
#     # Concatenate all embeddings
#     all_embeddings = torch.cat(all_embeddings_batches)
#
#     subword_indices = []
#     for start_boundary, end_boundary in boundaries_list:
#         subword_indices += list(range(start_boundary, end_boundary))
#
#     return subword_indices, all_embeddings
# def average_embeddings_from_overlapping_sliding_windows(tokens_dict, subword_indices, all_embeddings):
#     # Step 1: Collect embeddings and sum them directly
#     for token in tokens_dict:
#         token['embedding_sum'] = None
#         token['embedding_count'] = 0
#
#     # Step 2: Sum embeddings in place
#     for embedding_index, embedding in zip(subword_indices, all_embeddings):
#         token = tokens_dict[embedding_index]
#
#         # Initialize sum if not set
#         if token['embedding_sum'] is None:
#             token['embedding_sum'] = embedding
#         else:
#             token['embedding_sum'] += embedding
#
#         token['embedding_count'] += 1
#
#     # Step 3: Calculate average embeddings
#     for token in tqdm(tokens_dict, total=len(tokens_dict), desc="Calculating average embeddings", leave=False):
#         if token['embedding_count'] > 0:
#             token['average_embedding'] = token['embedding_sum'] / token['embedding_count']
#
#         # Clean up
#         del token['embedding_sum']
#         del token['embedding_count']
#
#     return tokens_dict
# def get_token_embeddings_tensor_from_subwords(tokens_df, tokens_dict, hidden_size, first_last_average=True):
#     ## Pre-allocate zero tensor to avoid creating it in each iteration
#     zero_tensor = torch.zeros(hidden_size)
#
#     # Pre-build the dictionary for collecting embeddings
#     tokens_df_embeddings_dict = {token_id: [] for token_id in tokens_df['token_ID_within_document'].tolist()}
#
#     # Populate the embeddings dictionary from tokens_dict
#     for token in tokens_dict:
#         token_id = token['token_id']
#         tokens_df_embeddings_dict[token_id].append(token['average_embedding'])
#
#     tokens_embeddings = []
#
#     # Process each token's subword embeddings
#     for token_id in tqdm(tokens_df_embeddings_dict.keys(), desc='Averaging subwords embeddings', leave=False):
#         embeddings = tokens_df_embeddings_dict[token_id]
#
#         if len(embeddings) == 0:
#             embeddings = [zero_tensor]  # Use pre-allocated zero tensor
#
#         if first_last_average:
#             # Take only the first and last embedding
#             first_last = torch.stack([embeddings[0], embeddings[-1]], dim=0)
#             average_embedding = torch.mean(first_last, dim=0)
#         else:
#             # Use all embeddings
#             average_embedding = torch.mean(torch.stack(embeddings, dim=0), dim=0)
#
#         tokens_embeddings.append(average_embedding)
#
#     # Stack the list of embeddings into a tensor
#     tokens_embeddings_tensor = torch.stack(tokens_embeddings)
#
#     return tokens_embeddings_tensor
#
# def load_tokenizer_and_embedding_model(model_name="almanach/camembert-base"):
#
#     tokenizer = AutoTokenizer.from_pretrained(model_name)
#     model = AutoModel.from_pretrained(model_name, output_hidden_states=True)
#     print(f"Tokenizer and Embedding Model Initialized: {model_name}")
#
#     device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
#     model.to(device)
#
#     return tokenizer, model
# def get_embedding_tensor_from_tokens_df(tokens_df, tokenizer, model, sliding_window_size='max', mini_batch_size=10,
#                                         sliding_window_overlap=0.5, first_last_average=True, device=None):
#     if device is None:
#         device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
#
#     # model_max_length = model.config.max_position_embeddings
#     model_max_length = tokenizer.model_max_length
#     hidden_size = model.config.hidden_size
#
#     if sliding_window_size == "max":
#         sliding_window_size = model_max_length
#
#     padding_token_id = int(tokenizer.pad_token_id)
#     tokens_dict = fast_tokennize_tokens_df(tokens_df, tokenizer)
#     boundaries_list = get_boudaries_list(max_token_id=len(tokens_dict), sliding_window_size=sliding_window_size,
#                                          sliding_window_overlap=sliding_window_overlap)
#     subword_indices, all_embeddings = compute_sub_word_embeddings(boundaries_list, tokens_dict, model,
#                                                                   mini_batch_size=mini_batch_size,
#                                                                   padding_token_id=padding_token_id,
#                                                                   sliding_window_size=sliding_window_size,
#                                                                   device=device)
#     tokens_dict = average_embeddings_from_overlapping_sliding_windows(tokens_dict, subword_indices, all_embeddings)
#     tokens_embeddings_tensor = get_token_embeddings_tensor_from_subwords(tokens_df, tokens_dict, hidden_size,
#                                                                          first_last_average=first_last_average)
#
#     return tokens_embeddings_tensor



from tqdm.auto import tqdm
import os

def reconstruct_text_from_tokens_df(tokens_df):
    """
    Reconstructs the original text from a DataFrame of tokens, preserving paragraph breaks
    and ensuring proper alignment with byte offsets.

    Args:
        tokens_df (DataFrame): DataFrame containing the tokenized text with columns:
                               'paragraph_ID', 'word', 'byte_onset', 'byte_offset'.

    Returns:
        str: Reconstructed text as a single string with paragraph breaks.
    """
    current_char_id = 0
    current_paragraph_ID = 0
    generated_text = ""

    for paragraph_ID, word, byte_onset, byte_offset in tokens_df[
        ["paragraph_ID", "word", "byte_onset", "byte_offset"]].values:
        word = str(word)

        # Add a newline for each new paragraph
        if paragraph_ID > current_paragraph_ID:
            generated_text += "\n"
            current_char_id += 1
            current_paragraph_ID = paragraph_ID

        # Add spaces to align with the byte offset
        while current_char_id < byte_onset:
            generated_text += " "
            current_char_id += 1

        # Append the word and update the current character ID
        generated_text += word
        current_char_id = byte_offset

    return generated_text

def generating_sacr_annotations(raw_text,
                                tokens_df,
                                entities_df,
                                entity_type_column="cat",
                                coref_name_column="COREF"):
    """
    Generates SACR-compatible annotations by embedding inline entity mentions and coreference
    chains within the raw text.

    Args:
        raw_text (str): The original reconstructed text.
        tokens_df (DataFrame): DataFrame with tokenized text, including byte offsets.
        entities_df (DataFrame): DataFrame with entity mentions and their metadata, including:
                                 'start_token', 'end_token', 'mention_len', and the columns
                                 specified by `entity_type_column` and `coref_name_column`.
        entity_type_column (str): Column name for entity type (default: "cat").
        coref_name_column (str): Column name for coreference ID (default: "COREF").

    Returns:
        str: Annotated text with inline SACR-compatible annotations.
    """
    annotated_text = raw_text

    # Sort entities by length of mention and start position for proper insertion order
    entities_df = entities_df.sort_values(by=['mention_len', 'start_token'], ascending=[True, True]).reset_index(
        drop=True)

    for COREF_name, start_token, end_token, cat, text in tqdm(
            entities_df[[coref_name_column, 'start_token', 'end_token', entity_type_column, 'text']].values,
            desc="Generating Sacr Annotations"):
        # Retrieve byte offsets for the mention
        origin_mention_byte_onset = tokens_df.loc[start_token, 'byte_onset']
        origin_mention_byte_offset = tokens_df.loc[end_token, 'byte_offset']

        # Construct the inline annotation
        mention_prefix = f'{{{COREF_name}:EN="{cat}" '
        mention_suffix = '}'
        annotated_entity = mention_prefix + annotated_text[
                                            origin_mention_byte_onset:origin_mention_byte_offset] + mention_suffix

        # Adjust token offsets to account for annotation length changes
        tokens_df["temp_byte_onset"], tokens_df["temp_byte_offset"] = tokens_df["byte_onset"], tokens_df["byte_offset"]

        # Update byte_onset and byte_offset for tokens after the annotation insertion point
        byte_onset_to_change = tokens_df[tokens_df["byte_onset"] > origin_mention_byte_onset]
        tokens_df.loc[byte_onset_to_change.index, 'temp_byte_onset'] = byte_onset_to_change['temp_byte_onset'] + len(
            mention_prefix)

        byte_onset_to_change = tokens_df[tokens_df["byte_onset"] >= origin_mention_byte_onset]
        tokens_df.loc[byte_onset_to_change.index, 'temp_byte_offset'] = byte_onset_to_change['temp_byte_offset'] + len(
            mention_prefix)

        byte_onset_to_change = tokens_df[tokens_df["byte_offset"] >= origin_mention_byte_offset]
        tokens_df.loc[byte_onset_to_change.index, 'temp_byte_offset'] = byte_onset_to_change['temp_byte_offset'] + len(
            mention_suffix)

        byte_onset_to_change = tokens_df[tokens_df["byte_offset"] > origin_mention_byte_offset]
        tokens_df.loc[byte_onset_to_change.index, 'temp_byte_onset'] = byte_onset_to_change['temp_byte_onset'] + len(
            mention_suffix)

        # Apply the updated byte offsets
        tokens_df["byte_onset"], tokens_df["byte_offset"] = tokens_df["temp_byte_onset"], tokens_df["temp_byte_offset"]

        # Insert the annotated entity into the text
        annotated_text = annotated_text[:origin_mention_byte_onset] + annotated_entity + annotated_text[
                                                                                         origin_mention_byte_offset:]

    return annotated_text

def generate_sacr_file(file_name,
                       tokens_df,
                       entities_df,
                       end_directory,
                       entity_type_column="cat",
                       coref_name_column="COREF",
                       sacr_extension=".generated_sacr"):
    """
    Generates a SACR file with inline annotations for mention detection and coreference chains.

    Args:
        file_name (str): Base name for the output file (without extension).
        tokens_df (DataFrame): DataFrame with tokenized text, including byte offsets.
        entities_df (DataFrame): DataFrame with entity metadata.
        end_directory (str): Directory to save the generated SACR file.
        entity_type_column (str): Column name for entity type (default: "cat").
        coref_name_column (str): Column name for coreference ID (default: "COREF").
        sacr_extension (str): File extension for the SACR file (default: ".generated_sacr").

    Returns:
        None: Saves the annotated SACR file to the specified directory.
    """
    # Reconstruct the original raw text from tokens
    raw_text = reconstruct_text_from_tokens_df(tokens_df)

    # Generate the SACR-compatible annotated text
    sacr_annotated_text = generating_sacr_annotations(raw_text,
                                                      tokens_df,
                                                      entities_df,
                                                      entity_type_column=entity_type_column,
                                                      coref_name_column=coref_name_column)

    import re
    # Add additional spacing for paragraph formatting
    sacr_annotated_text = sacr_annotated_text.replace("\n", "\n\n")
    # Replace multiple spaces (but not newlines) with a single space
    sacr_annotated_text = re.sub(r'(?<=\S) {2,}(?=\S)', ' ', sacr_annotated_text)
    sacr_annotated_text = sacr_annotated_text.replace("\n\n ", "\n\n")
    sacr_annotated_text = sacr_annotated_text.replace("\n ", "\n")
    sacr_annotated_text = sacr_annotated_text.replace("\n ", "\n")

    # Save the generated SACR file
    file_path = os.path.join(end_directory, f"{file_name}{sacr_extension}")
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(sacr_annotated_text)

    print(f"File saved at:\n{file_path}")

    return sacr_annotated_text